subroutine CalcEnergy(Pos, M, L, rc, lambdaD, epsilonC, PEnergy, qList, Dim, NAtom)
    implicit none
    integer, intent(in) :: M, Dim, NAtom
    real(8), intent(in), dimension(0:NAtom-1, 0:Dim-1) :: Pos
    ! initialize list of particle charges
    real(8), intent(in), dimension(0:NAtom-1) :: qList
    ! initialize box length, cutoff radius, Debye length, and Coulomb prefactor
    real(8), intent(in) :: L, rc, lambdaD, epsilonC
    real(8), intent(out) :: PEnergy
    real(8), parameter :: k = 3000., r0 = 1.
    real(8), dimension(Dim) :: rij, Posi
    real(8) :: irc6, rc2, d, d2, id2, id6, id12, ilD, Shift, ShiftC, deltarij, qi, qij, CoulEnergy
    ! initialize loop counters
    integer :: i, j
    
    PEnergy = 0.
    ! pre-calculate rc**2
    rc2 = rc*rc
    ! calculate 1/rc**6
    irc6 = 1./(rc2*rc2*rc2)
    ! pre-calculate 1/lambdaD
    ilD = 1./lambdaD
    ! calculate LJ potential shift using pre-calculated value of irc6
    Shift = -4. * (irc6*irc6 - irc6)
    ! calculate magnitude of Coulomb potential shift
    ShiftC = epsilonC*exp(-rc*ilD)/rc
    
    ! pairwise loop for all unique i, j
    do i = 0, NAtom - 1
        ! store the ith row of position vector in a temp array
        Posi = Pos(i,:)
        ! store the ith charge
        qi = qList(i)
        do j = i + 1, NAtom - 1
            ! calculate vector distance from i to j
            rij = Pos(j,:) - Posi
            ! calculate image distance
            rij = rij - L * dnint(rij/L)
            ! compute squared magnitude for comparison to rc
            d2 = sum(rij*rij)
            ! ...if i, j are bonded, then use harmonic potential
            if (j == i + 1 .and. mod(j, M) > 0) then
                d = sqrt(d2)
                deltarij = d - r0
                ! calculate energy
                PEnergy = PEnergy + (k/2.) * deltarij*deltarij
            ! ...if i, j are not bonded, then use shifted LJ and Coulomb
            else
                ! end this iteration if j outside cutoff
                if (d2 > rc2) then
                    cycle
                endif
                ! pre-calculate inverse distances for energy and force calcs
                d = sqrt(d2)
                id2 = 1./d2
                id6 = id2*id2*id2
                id12 = id6*id6
                ! pre-calculate qi*qj and exponential term
                qij = qi * qList(j)
                CoulEnergy = epsilonC*exp(-d*ilD)/d
                ! calculate energy due to LJ and Coulomb with shifts
                PEnergy = PEnergy + 4. * (id12 - id6) + Shift + qij*(CoulEnergy - ShiftC)
            endif

        enddo
    enddo
end subroutine



 subroutine CalcEnergyForces(Pos, M, L, rc, lambdaD, epsilonC, PEnergy, Forces, qList, Dim, NAtom)
     implicit none
     integer, intent(in) :: M, Dim, NAtom
     ! initialize position vector
     real(8), intent(in), dimension(0:NAtom-1, 0:Dim-1) :: Pos
    ! initialize list of particle charges
    real(8), intent(in), dimension(0:NAtom-1) :: qList
    ! initialize box length, cutoff radius, Debye length, and Coulomb prefactor
     real(8), intent(in) :: L, rc, lambdaD, epsilonC
     real(8), intent(out) :: PEnergy
     real(8), intent(inout), dimension(0:NAtom-1, 0:Dim-1) :: Forces
     !f2py intent(in, out) :: Forces
     real(8), parameter :: k = 3000., r0 = 1.
     real(8), dimension(Dim) :: rij, Fij, Posi
     real(8) :: irc6, rc2, d, d2, id1, id2, id6, id12, ilD, Shift, ShiftC, deltarij, qi, qij, CoulEnergy
     ! initialize loop counters
     integer :: i, j
     
     PEnergy = 0.
     Forces = 0.
    ! pre-calculate rc**2
    rc2 = rc*rc
    ! calculate 1/rc**6
    irc6 = 1./(rc2*rc2*rc2)
    ! pre-calculate 1/lambdaD
    ilD = 1./lambdaD
    ! calculate LJ potential shift using pre-calculated value of irc6
    Shift = -4. * (irc6*irc6 - irc6)
    ! calculate magnitude of Coulomb potential shift
    ShiftC = epsilonC*exp(-rc*ilD)/rc

     ! pairwise loop for all unique i, j
     do i = 0, NAtom - 1
         ! store the ith row of position vector in a temp array
         Posi = Pos(i,:)
        ! store the ith charge
        qi = qList(i)
         do j = i + 1, NAtom - 1
             ! calculate vector distance from i to j
             rij = Pos(j,:) - Posi
             ! calculate image distance
             rij = rij - L * dnint(rij/L)
             ! compute squared magnitude for comparison to rc
             d2 = sum(rij*rij)
             ! ...if i, j are bonded, then use harmonic potential
             if (j == i + 1 .and. mod(j, M) > 0) then
                 d = sqrt(d2)
                 deltarij = d - r0
                 ! calculate energy
                 PEnergy = PEnergy + (k/2.) * deltarij*deltarij
                 ! calculate forces; directions are given by unit vector of rij
                 Fij = k * deltarij * rij/d
             ! ...if i, j are not bonded, then use shifted LJ and Coulomb
             else
                 ! end this iteration if j outside cutoff
                 if (d2 > rc2) then
                     cycle
                 endif
                ! pre-calculate inverse distances for energy and force calcs
                d = sqrt(d2)
                id1 = 1./d
                id2 = 1./d2
                id6 = id2*id2*id2
                id12 = id6*id6
                ! pre-calculate qi*qj and exponential term
                qij = qi * qList(j)
                CoulEnergy = epsilonC*exp(-d*ilD)*id1
                 ! calculate energy with shift
                 PEnergy = PEnergy + 4. * (id12 - id6) + Shift + qij*(CoulEnergy - ShiftC)
                 ! calculate forces
                 Fij = rij * ((-48. * id12 + 24. * id6) * id2 + qij*CoulEnergy*(id1 - ilD)*id1)
             endif
             
             ! store forces for i and j
             Forces(i,:) = Forces(i,:) + Fij
             Forces(j,:) = Forces(j,:) - Fij
         enddo
     enddo
     
 end subroutine
 
 
 
 subroutine CalcEnergyForcesRDF(Pos, M, L, rc, lambdaD, epsilonC, PEnergy, Forces, qList, AtomCoord, Dim, NAtom)
     implicit none
     integer, intent(in) :: M, Dim, NAtom
     ! initialize position vector
     real(8), intent(in), dimension(0:NAtom-1, 0:Dim-1) :: Pos
    ! initialize list of particle charges
    real(8), intent(in), dimension(0:NAtom-1) :: qList
    ! initialize box length, cutoff radius, Debye length, and Coulomb prefactor
     real(8), intent(in) :: L, rc, lambdaD, epsilonC
     real(8), intent(out) :: PEnergy
     real(8), intent(inout), dimension(0:NAtom-1, 0:Dim-1) :: Forces
     !f2py intent(in, out) :: Forces
     real(8), intent(out), dimension(0:NAtom-1) :: AtomCoord ! coordination numbers
     real(8), parameter :: k = 3000., r0 = 1., CoordCutoff = 1.5
     real(8), dimension(Dim) :: rij, Fij, Posi
     real(8) :: irc6, rc2, d, d2, id1, id2, id6, id12, ilD, Shift, ShiftC, deltarij, qi, qij, CoulEnergy
     ! initialize loop counters and RDF bin index
     integer :: i, j
     
     PEnergy = 0.
     Forces = 0.
     AtomCoord = 0.
     
    ! pre-calculate rc**2
    rc2 = rc*rc
    ! calculate 1/rc**6
    irc6 = 1./(rc2*rc2*rc2)
    ! pre-calculate 1/lambdaD
    ilD = 1./lambdaD
    ! calculate LJ potential shift using pre-calculated value of irc6
    Shift = -4. * (irc6*irc6 - irc6)
    ! calculate magnitude of Coulomb potential shift
    ShiftC = epsilonC*exp(-rc*ilD)/rc

     ! pairwise loop for all unique i, j
     do i = 0, NAtom - 1
         ! store the ith row of position vector in a temp array
         Posi = Pos(i,:)
        ! store the ith charge
        qi = qList(i)
         do j = i + 1, NAtom - 1
             ! calculate vector distance from i to j
             rij = Pos(j,:) - Posi
             ! calculate image distance
             rij = rij - L * dnint(rij/L)
             ! compute squared magnitude for comparison to rc
             d2 = sum(rij*rij)
             ! ...if i, j are bonded, then use harmonic potential
             if (j == i + 1 .and. mod(j, M) > 0) then
                 d = sqrt(d2)
                 deltarij = d - r0
                 ! calculate energy
                 PEnergy = PEnergy + (k/2.) * deltarij*deltarij
                 ! calculate forces; directions are given by unit vector of rij
                 Fij = k * deltarij * rij/d
             ! ...if i, j are not bonded, then use shifted LJ and Coulomb
             else
                 ! end this iteration if j outside cutoff
                 if (d2 > rc2) then
                     cycle
                 endif
                ! pre-calculate inverse distances for energy and force calcs
                d = sqrt(d2)
                id1 = 1./d
                id2 = 1./d2
                id6 = id2*id2*id2
                id12 = id6*id6
                ! pre-calculate qi*qj and exponential term
                qij = qi * qList(j)
                CoulEnergy = epsilonC*exp(-d*ilD)*id1
                 ! calculate energy with shift
                 PEnergy = PEnergy + 4. * (id12 - id6) + Shift + qij*(CoulEnergy - ShiftC)
                 ! calculate forces
                 Fij = rij * ((-48. * id12 + 24. * id6) * id2 + qij*CoulEnergy*(id1 - ilD)*id1)
             endif
             ! store forces for i and j
             Forces(i,:) = Forces(i,:) + Fij
             Forces(j,:) = Forces(j,:) - Fij
             ! count atoms within CoordCutoff and add tally to AtomCoord
             if (d <= CoordCutoff) then
                 AtomCoord(i) = AtomCoord(i) + 1
                 AtomCoord(j) = AtomCoord(j) + 1
            endif
         enddo
     enddo
     
 end subroutine



subroutine VVIntegrate(Pos, Vel, Accel, M, L, rc, lambdaD, epsilonC, dt, KEnergy, PEnergy, qList, Dim, NAtom)
    implicit none
    integer, intent(in) :: M, Dim, NAtom
    real(8), intent(in) :: L, rc, dt, lambdaD, epsilonC
    real(8), intent(inout), dimension(0:NAtom-1, 0:Dim-1) :: Pos, Vel, Accel
!f2py intent(in,out) :: Pos, Vel, Accel
    ! initialize list of particle charges
    real(8), intent(in), dimension(0:NAtom-1) :: qList
    real(8), intent(out) :: KEnergy, PEnergy
    external :: CalcEnergyForces
    Pos = Pos + dt * Vel + 0.5 * dt*dt * Accel
    Vel = Vel + 0.5 * dt * Accel
    call CalcEnergyForces(Pos, M, L, rc, lambdaD, epsilonC, PEnergy, Accel, qList, Dim, NAtom)
    Vel = Vel + 0.5 * dt * Accel
    KEnergy = 0.5 * sum(Vel*Vel)
end subroutine



subroutine VVIntegrateRDF(Pos, Vel, Accel, M, L, rc, lambdaD, epsilonC, dt, KEnergy, PEnergy, qList, AtomCoord, Dim, NAtom)
    implicit none
    integer, intent(in) :: M, Dim, NAtom
    real(8), intent(in) :: L, rc, dt, lambdaD, epsilonC
    real(8), intent(inout), dimension(0:NAtom-1, 0:Dim-1) :: Pos, Vel, Accel
!f2py intent(in,out) :: Pos, Vel, Accel, g
    ! initialize list of particle charges
    real(8), intent(in), dimension(0:NAtom-1) :: qList
    real(8), intent(out) :: KEnergy, PEnergy
    real(8), intent(out), dimension(0:NAtom-1) :: AtomCoord ! coordination numbers
    external :: CalcEnergyForcesRDF
    Pos = Pos + dt * Vel + 0.5 * dt*dt * Accel
    Vel = Vel + 0.5 * dt * Accel
    call CalcEnergyForcesRDF(Pos, M, L, rc, lambdaD, epsilonC, PEnergy, Accel, qList, AtomCoord, Dim, NAtom)
    Vel = Vel + 0.5 * dt * Accel
    KEnergy = 0.5 * sum(Vel*Vel)
end subroutine




!subroutine CalcEnergyVirial(Pos, M, L, rc, PEnergy, W, lambdaD, epsilonS, Dim, NAtom)
!    implicit none
!    integer, intent(in) :: M, Dim, NAtom
!    real(8), intent(in), dimension(0:NAtom-1, 0:Dim-1) :: Pos
!    real(8), intent(in) :: L, rc, lambdaD, epsilonS
!    real(8), intent(out) :: PEnergy, W
!    real(8), parameter :: k = 3000., r0 = 1.
!    !======== YOUR CODE HERE ========
!    real(8), dimension(Dim) :: rij, Posi
!    real(8) :: irc6, rc2, d, d2, id2, id6, id12, ilD, Shift, ShiftC, deltarij
!    ! initialize loop counters
!    integer :: i, j
!    
!    PEnergy = 0.
!    W = 0.
!    ! pre-calculate rc**2
!    rc2 = rc*rc
!    ! pre-calculate 1/lambdaD
!    ilD = 1./lambdaD
!    ! calculate 1/rc**6
!    irc6 = 1./(rc2*rc2*rc2)
!    ! calculate LJ potential shift using pre-calculated value of irc6
!    Shift = -4. * (irc6*irc6 - irc6)
!    ! calculate Coulomb potential shift
!    ShiftC = 0
!    
!    
!    ! pairwise loop for all unique i, j
!    do i = 0, NAtom - 1
!        ! store the ith row of position vector in a temp array
!        Posi = Pos(i,:)
!        do j = i + 1, NAtom - 1
!            ! calculate vector distance from i to j
!            rij = Pos(j,:) - Posi
!            ! calculate image distance
!            rij = rij - L * dnint(rij/L)
!            ! compute squared magnitude for comparison to rc
!            d2 = sum(rij*rij)
!            ! ...if i, j are bonded, then use harmonic potential
!            if (j == i + 1 .and. mod(j, M) > 0) then
!                d = sqrt(d2)
!                deltarij = d - r0
!                ! calculate energy
!                PEnergy = PEnergy + (k/2.) * deltarij*deltarij
!               ! calculate virial contribution
!                W = W + k * deltarij * d
!            ! ...if i, j are BOTH CHARGED, then use screened Coulomb
!            ! since charged beads occur every 3 atoms
!            if (mod(i, 3) == 0 .and. mod(j, 3) == 0) then
!                d = sqrt(d2)
!                ! repulsive interaction: i, j have same signs
!                if ((mod(i, 2) == 0 .and. mod(j, 2) == 0) .or. (mod(i, 2) ~= 0 .and. mod(j, 2) ~= 0)) then
!                    PEnergy = PEnergy + 1./(4.*pi*epsilonS*d) * exp(-d*ilD)
!                ! attractive interaction: i, j have different signs
!                else
!                    PEnergy = PEnergy - 1./(4.*pi*epsilonS*d) * exp(-d*ilD)
!                endif ! NEED TO ADD VIRIAL!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!            ! ...if i, j are not bonded, then use shifted LJ
!            else
!                ! end this iteration if j outside cutoff
!                if (d2 > rc2) then
!                    cycle
!                endif
!                ! pre-calculate inverse distances for energy and force calcs
!                id2 = 1./d2
!                id6 = id2*id2*id2
!                id12 = id6*id6
!                ! calculate energy with shift
!                PEnergy = PEnergy + 4. * (id12 - id6) + Shift
!                ! calculate virial contribution
!                W = W - 48. * id12 + 24. * id6
!            endif
!
!        enddo
!    enddo
!end subroutine


